use color_eyre::{eyre::ContextCompat, Result};
use tiny_skia::{ColorU8, Pixmap, PremultipliedColorU8};

use crate::model::image_filter::ImageFilter;
use crate::model::rgb::RGB8Wrapper;

pub fn apply(original: &Pixmap, filter: &ImageFilter) -> Result<Pixmap> {
    let mut result = if let Some(rect) = filter.crop {
        original.clone_rect(*rect).wrap_err_with(|| format!("Invalid crop rect: {}", rect))?
    } else {
        original.clone()
    };

    // scale is handled in runner::graphics::render_key

    // rotate

    if let Some(color) = &filter.color {
        apply_color(&mut result, color);
    }

    if filter.alpha != 1.0 {
        apply_alpha(&mut result, filter.alpha);
    }

    // grayscale

    if filter.invert {
        apply_invert(&mut result);
    }

    Ok(result)
}

fn apply_alpha(pixmap: &mut Pixmap, alpha: f32) {
    for p in pixmap.pixels_mut() {
        let d = p.demultiply();
        *p = ColorU8::from_rgba(d.red(), d.green(), d.blue(), (d.alpha() as f32 * alpha).round() as u8).premultiply();
    }
}

fn apply_invert(pixmap: &mut Pixmap) {
    for p in pixmap.pixels_mut() {
        *p = PremultipliedColorU8::from_rgba(p.alpha() - p.red(), p.alpha() - p.green(), p.alpha() - p.blue(), p.alpha()).unwrap();
    }
}

fn apply_color(pixmap: &mut Pixmap, color: &RGB8Wrapper) {
    for p in pixmap.pixels_mut() {
        let a = p.alpha();
        if a > 0 {
            *p = ColorU8::from_rgba(color.r, color.g, color.b, a).premultiply();
        }
    }
}
