use std::collections::hash_map::Entry;
use std::collections::{HashMap, HashSet};
use std::path::Path;

use color_eyre::eyre::{eyre, ContextCompat, WrapErr};
use color_eyre::Result;
use resvg::usvg::{TextRendering, TreeParsing, TreeTextToPath};
use tiny_skia::{Pixmap, Transform};

use crate::model::config::{Config, IconFormat, IconPack};
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::model::IconMap;

mod filter;

#[derive(Debug)]
pub struct LoadedIcon {
    pub pixmap: Pixmap,
    pub scale: f32,
}

pub fn get_used_icon_descriptors(config: &Config) -> HashSet<IconDescriptor> {
    let mut result: HashSet<IconDescriptor> = HashSet::new();

    fn insert_all_from_map<T>(result: &mut HashSet<IconDescriptor>, map: &IconMap<T>) {
        map.values().for_each(|v| {
            result.insert(v.clone());
        });
    }

    for page in config.key_pages_by_id.values() {
        for key in page.keys.values() {
            if let Some(d) = &key.icon {
                result.insert(d.clone());

                if let Some(c) = &key.mode.media__next {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.media__play_pause {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.media__previous {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.home_assistant__button {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.home_assistant__switch {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.spotify__repeat {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.spotify__shuffle {
                    insert_all_from_map(&mut result, &c.icon);
                }
            }
        }
    }

    for page in config.knob_pages_by_id.values() {
        for knob in page.knobs.values() {
            result.insert(knob.icon.clone());

            if let Some(c) = &knob.mode.audio_volume {
                insert_all_from_map(&mut result, &c.icon);
            }
        }
    }

    result
}

pub fn load_icons(
    config_directory: &Path,
    icon_packs_by_id: &HashMap<String, IconPack>,
    descriptors: HashSet<IconDescriptor>,
    dpi: f32,
) -> Result<HashMap<IconDescriptor, LoadedIcon>> {
    let mut unfiltered_pixmap_by_source: HashMap<IconDescriptorSource, Pixmap> = HashMap::new();
    let mut icons_by_descriptor: HashMap<IconDescriptor, LoadedIcon> = HashMap::new();
    let mut fonts_db = resvg::usvg::fontdb::Database::new();
    fonts_db.load_system_fonts();

    for descriptor in descriptors {
        if descriptor.source == IconDescriptorSource::None {
            continue;
        }

        let original_image = match unfiltered_pixmap_by_source.entry(descriptor.source.clone()) {
            Entry::Occupied(o) => o.into_mut(),
            Entry::Vacant(v) => v.insert(read_image(config_directory, icon_packs_by_id, dpi, &fonts_db, descriptor.source.clone())?),
        };

        let pixmap = filter::apply(original_image, &descriptor.filter)?;
        let scale = descriptor.filter.scale;

        icons_by_descriptor.insert(descriptor, LoadedIcon { pixmap, scale });
    }

    Ok(icons_by_descriptor)
}

fn read_image(
    config_directory: &Path,
    icon_packs_by_id: &HashMap<String, IconPack>,
    dpi: f32,
    fonts_db: &resvg::usvg::fontdb::Database,
    source: IconDescriptorSource,
) -> Result<Pixmap> {
    let path = match source {
        IconDescriptorSource::None => return Ok(Pixmap::new(1, 1).unwrap()),
        IconDescriptorSource::Path(path) => path,
        IconDescriptorSource::IconPack { pack_id, icon_id } => {
            let pack = icon_packs_by_id.get(&pack_id).wrap_err_with(|| format!("Unknown icon pack: @{}", pack_id))?;

            let extension = match pack.format {
                IconFormat::Png => "png",
                IconFormat::Svg => "svg",
            };

            config_directory.join(&pack.path).join(icon_id + "." + extension)
        }
    };

    Ok(match path.extension() {
        None => return Err(eyre!("Invalid icon path: {:?}", path)),
        Some(extension) => match extension.to_string_lossy().as_ref() {
            "png" => Pixmap::load_png(&path).wrap_err_with(|| format!("Failed to open or decode the PNG file at {}", path.to_string_lossy()))?,
            "svg" => {
                read_image_from_svg(&path, dpi, fonts_db).wrap_err_with(|| format!("Failed to open or decode the SVG file at {}", path.to_string_lossy()))?
            }
            extension => return Err(eyre!("Invalid file extension, only *.png and *.svg are allowed: {}", extension)),
        },
    })
}

fn read_image_from_svg(path: &Path, dpi: f32, font_db: &resvg::usvg::fontdb::Database) -> Result<Pixmap> {
    let raw_data = std::fs::read(path)?;

    let tree = {
        let mut tree = resvg::usvg::Tree::from_data(
            &raw_data,
            &resvg::usvg::Options {
                dpi,
                font_family: "Inter".to_owned(),
                font_size: 11.0,
                text_rendering: TextRendering::OptimizeLegibility,
                ..resvg::usvg::Options::default()
            },
        )?;

        tree.convert_text(font_db);

        resvg::Tree::from_usvg(&tree)
    };

    let size = tree.size.to_int_size();
    let mut pixmap = Pixmap::new(size.width(), size.height()).unwrap();

    tree.render(Transform::default(), &mut pixmap.as_mut());

    Ok(pixmap)
}
