use std::collections::HashMap;

use bytes::{BufMut, Bytes, BytesMut};
use tiny_skia::{BlendMode, Color, FilterQuality, IntSize, Pixmap, PixmapPaint, PremultipliedColorU8, Transform};

use loupedeck_serial::util::Endianness;

use crate::icons::LoadedIcon;
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::runner::state::Key;

pub fn render_key(key_size: IntSize, buffer_endianness: Endianness, icons: &HashMap<IconDescriptor, LoadedIcon>, state: Option<&Key>) -> Bytes {
    let mut canvas = Pixmap::new(key_size.width(), key_size.height()).unwrap();

    if let Some(state) = state {
        if state.icon.source != IconDescriptorSource::None {
            let icon = &icons[&state.icon];
            let scaled_size = IntSize::from_wh(icon.pixmap.width(), icon.pixmap.height())
                .unwrap()
                .scale_by(icon.scale)
                .unwrap();

            static PAINT: PixmapPaint = PixmapPaint {
                opacity: 1.0,
                blend_mode: BlendMode::SourceOver,
                quality: FilterQuality::Bicubic,
            };

            canvas.draw_pixmap(
                (((key_size.width() - scaled_size.width()) / 2) as f32 / icon.scale).round() as i32,
                (((key_size.height() - scaled_size.height()) / 2) as f32 / icon.scale).round() as i32,
                icon.pixmap.as_ref(),
                &PAINT,
                Transform::from_scale(icon.scale, icon.scale),
                None,
            );
        }
    } else {
        canvas.fill(Color::BLACK);
    }

    convert_pixels_to_rgb565(canvas.pixels(), buffer_endianness).freeze()
}

fn convert_pixels_to_rgb565(pixels: &[PremultipliedColorU8], endianness: Endianness) -> BytesMut {
    let pixel_count = pixels.len();

    let mut result = BytesMut::with_capacity(pixel_count * 2);

    for pixel in pixels {
        let red = pixel.red() as u16;
        let green = pixel.green() as u16;
        let blue = pixel.blue() as u16;
        let color = ((red & 0b11111000) << 8) | ((green & 0b11111100) << 3) | (blue.wrapping_shr(3));

        match endianness {
            Endianness::LittleEndian => result.put_u16_le(color),
            Endianness::BigEndian => result.put_u16(color),
        }
    }

    result
}
